@@ -115,7 +115,7 @@ typedef struct XMVDemuxContext {
     XMVAudioPacket *audio; ///< The audio packets contained in each packet.
 } XMVDemuxContext;
 
-static int xmv_probe(AVProbeData *p)
+static int xmv_probe(const AVProbeData *p)
 {
     uint32_t file_version;
 
@@ -149,7 +149,6 @@ static int xmv_read_header(AVFormatContext *s)
     uint32_t file_version;
     uint32_t this_packet_size;
     uint16_t audio_track;
-    int ret;
 
     s->ctx_flags |= AVFMTCTX_NOHEADER;
 
@@ -176,11 +175,9 @@ static int xmv_read_header(AVFormatContext *s)
 
     avio_skip(pb, 2); /* Unknown (padding?) */
 
-    xmv->audio = av_mallocz_array(xmv->audio_track_count, sizeof(XMVAudioPacket));
-    if (!xmv->audio) {
-        ret = AVERROR(ENOMEM);
-        goto fail;
-    }
+    xmv->audio = av_calloc(xmv->audio_track_count, sizeof(*xmv->audio));
+    if (!xmv->audio)
+        return AVERROR(ENOMEM);
 
     for (audio_track = 0; audio_track < xmv->audio_track_count; audio_track++) {
         XMVAudioPacket *packet = &xmv->audio[audio_track];
@@ -214,8 +211,7 @@ static int xmv_read_header(AVFormatContext *s)
              packet->channels >= UINT16_MAX / XMV_BLOCK_ALIGN_SIZE) {
             av_log(s, AV_LOG_ERROR, "Invalid parameters for audio track %"PRIu16".\n",
                    audio_track);
-            ret = AVERROR_INVALIDDATA;
-            goto fail;
+            return AVERROR_INVALIDDATA;
         }
     }
 
@@ -223,16 +219,10 @@ static int xmv_read_header(AVFormatContext *s)
     /* Initialize the packet context */
 
     xmv->next_packet_offset = avio_tell(pb);
-    if (this_packet_size < xmv->next_packet_offset)
-        return AVERROR_INVALIDDATA;
     xmv->next_packet_size   = this_packet_size - xmv->next_packet_offset;
     xmv->stream_count       = xmv->audio_track_count + 1;
 
     return 0;
-
-fail:
-    xmv_read_close(s);
-    return ret;
 }
 
 static void xmv_read_extradata(uint8_t *extradata, AVIOContext *pb)
@@ -342,7 +332,7 @@ static int xmv_process_packet_header(AVFormatContext *s)
             ast->codecpar->codec_type            = AVMEDIA_TYPE_AUDIO;
             ast->codecpar->codec_id              = packet->codec_id;
             ast->codecpar->codec_tag             = packet->compression;
-            ast->codecpar->channels              = packet->channels;
+            ast->codecpar->ch_layout.nb_channels = packet->channels;
             ast->codecpar->sample_rate           = packet->sample_rate;
             ast->codecpar->bits_per_coded_sample = packet->bits_per_sample;
             ast->codecpar->bit_rate              = packet->bit_rate;
@@ -399,8 +389,6 @@ static int xmv_process_packet_header(AVFormatContext *s)
                 av_assert0(xmv->video.stream_index < s->nb_streams);
 
                 if (vst->codecpar->extradata_size < 4) {
-                    av_freep(&vst->codecpar->extradata);
-
                     if ((ret = ff_alloc_extradata(vst->codecpar, 4)) < 0)
                         return ret;
                 }
@@ -587,11 +575,12 @@ static int xmv_read_packet(AVFormatContext *s,
     return 0;
 }
 
-AVInputFormat ff_xmv_demuxer = {
+const AVInputFormat ff_xmv_demuxer = {
     .name           = "xmv",
     .long_name      = NULL_IF_CONFIG_SMALL("Microsoft XMV"),
     .extensions     = "xmv",
     .priv_data_size = sizeof(XMVDemuxContext),
+    .flags_internal = FF_FMT_INIT_CLEANUP,
     .read_probe     = xmv_probe,
     .read_header    = xmv_read_header,
     .read_packet    = xmv_read_packet,
